/**********************************************-*- indent-tabs-mode:nil; -*-
 *                                                                         *
 *   InConcert - Tap Tempo Synchronization for Live Perforances            *
 *                                                                         *
 *   Copyright (C) 2008 by Gabriel M. Beddingfield                         *
 *   gabriel@teuton.org                                                    *
 *                                                                         *
 *   "For of him [God], and through him, and unto him, are all things.     *
 *   To him be the glory for ever. Amen." (Romans 11:36)                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; version 2 of the License, or any later  *
 *   version.                                                              *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/* main.cpp
   Gabriel M. Beddingfield
   2008-02-10

   Driver to test different tap tempo models.

 */

#include <iostream>
#include <iomanip>
#include <deque>
#include <cmath>
#include <ctime>

#include <alsa/asoundlib.h>
#include <alsa/seq.h>

#include "TapTempoInterface.hpp"
#include "TapTempoAverage.hpp"
#include "TapTempoLeastSquares.hpp"
#include "TapTempoMedian.hpp"

using namespace std;

int main(void)
{
    snd_seq_t* seq;
    int seq_port;
    deque<TapTempoInterface*> tappers;
    deque<usecs_t> next_beat;
    deque<usecs_t> cycles_per_beat;
    deque<double> diffs;
    deque<double> diffsqrd;

    tappers.push_back(new TapTempoAverage);
    next_beat.push_back(0);
    cycles_per_beat.push_back(0);
    diffs.push_back(0.0);
    diffsqrd.push_back(0.0);
    tappers.push_back(new TapTempoLeastSquares);
    next_beat.push_back(0);
    cycles_per_beat.push_back(0);
    diffs.push_back(0.0);
    diffsqrd.push_back(0.0);
    tappers.push_back(new TapTempoMedian);
    next_beat.push_back(0);
    cycles_per_beat.push_back(0);
    diffs.push_back(0.0);
    diffsqrd.push_back(0.0);

    cout << "Tap Tempo Experiments" << endl
	 << "2008-02-10 Gabriel M. Beddingfield" << endl
	 << endl;

    cout << "Opening MIDI port" << endl;

    snd_seq_open(&seq, "default", SND_SEQ_OPEN_INPUT, 0);
    seq_port = snd_seq_create_simple_port(
	seq,
	"Tap Tempo Experiment In",
        SND_SEQ_PORT_CAP_WRITE|SND_SEQ_PORT_CAP_SUBS_WRITE,
        SND_SEQ_PORT_TYPE_MIDI_GENERIC|SND_SEQ_PORT_TYPE_APPLICATION
        |SND_SEQ_PORT_TYPE_SOFTWARE);

    cout << "Listening for taps on channel 12 (b)." << endl
	 << "Send 0xbb 0x32 0x7f (CC/12 50 127) to register a tap." << endl
	 << "Send 0xbb 0x31 0x00 (CC/12 49 0) to exit application."
	 << endl;

    const char* speeds[] = { "slowly", "medium", "fast" };
    const unsigned nspeeds =  3;
    const unsigned taps[] = { 3, 4, 5, 6 };
    const unsigned ntaps = 4;
    unsigned ntests = ntaps * nspeeds;
    unsigned test = 0;
    snd_seq_event_t* ev;
    usecs_t now;
    int j;
    bool not_done = true;
    while(not_done) {
	int getting_taps = taps[test/nspeeds];

	for(j=0; j<tappers.size(); ++j) {
	    tappers[j]->setTapsRequired(getting_taps-1);
	}	
	snd_seq_drop_input(seq);
	cout << "Please give " << getting_taps << " taps "
	     << speeds[test%nspeeds] << endl;

	while(getting_taps > 0) {
	    snd_seq_event_input(seq, &ev);
	    now = get_cycles();  // hurry!

	    // Filters
	    if(ev->type != SND_SEQ_EVENT_CONTROLLER)
		continue;
	    if(ev->data.control.channel != 0x0b)
		continue;
	    if(ev->data.control.param == 0x31)
		exit(0);
	    if((ev->data.control.param != 0x32)
	       && (ev->data.control.value & 0x40))
		continue;

	    cout << "Tap @ " << dec << now << endl;

	    // Register taps.
	    --getting_taps;
	    if(getting_taps == 0) continue;
	    for(j=0; j<tappers.size(); ++j) {
		next_beat[j] = tappers[j]->tap(now, cycles_per_beat[j]);
	    }
	}

	for(j=0; j<tappers.size(); ++j) {
            double diff;
            diff = (double(now)-double(next_beat[j]))/1000.0;
	    cout << tappers[j]->name() << ": "
		 << "cpb = " << dec << cycles_per_beat[j] << " "
		 << "next = " << dec << next_beat[j] << " "
		 << "diff = " << diff << " ms"
		 << endl;
            diffs[j] += fabs(diff);
            diffsqrd[j] += diff*diff;
	    next_beat[j] = 0;
	    cycles_per_beat[j] = 0;
	}

	++test;
	if(test >= ntests) not_done = false;
    }

    cout << "Final stats:" << endl;
    for(j=0; j<tappers.size(); ++j) {
        cout << tappers[j]->name() << ": "
             << diffs[j] << " ms / " 
             << diffsqrd[j] << " ms^2"
             << endl;
    }

    while(tappers.size() > 0) {
	delete tappers.front();
	tappers.pop_front();
    }

    cout << "Closing MIDI port" << endl;
    snd_seq_delete_simple_port(seq, seq_port);
    return 0;
}
