/**********************************************-*- indent-tabs-mode:nil; -*-
 *                                                                         *
 *   InConcert - Tap Tempo Synchronization for Live Performances           *
 *                                                                         *
 *   Copyright (C) 2008 by Gabriel M. Beddingfield                         *
 *   gabriel@teuton.org                                                    *
 *                                                                         *
 *   "For of him [God], and through him, and unto him, are all things.     *
 *   To him be the glory for ever. Amen." (Romans 11:36)                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; version 2 of the License, or any later  *
 *   version.                                                              *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/* MidiListener.hpp
   Gabriel M. Beddingfield
   2008-02-08

 */

#ifndef NDEBUG
#include <iostream>
#include <iomanip>
#endif

#ifdef USE_JACK_MIDI
#include <jack/jack.h>
#include <jack/midiport.h>
#else
#include <alsa/asoundlib.h>
#include <alsa/seq.h>
#endif

#include "MidiListener.hpp"
#include "TimeBase.hpp"

using namespace InConcert;

// Midi Implementation Limits
#define MIDI_CC_MIN  0x30
#define MIDI_CC_MAX  0x38
#define TAPS_MAX     12

// Midi Implementation
#define CC_SET_CHANNEL  0x30
#define CC_REZERO       0x31
#define CC_TAP          0x32
#define CC_STARTSTOP    0x33
#define CC_INSERTORSKIP 0x34
#define CC_TAPSREQUIRED 0x35
#define CC_TAPSCUECTL   0x36
#define CC_BEATSPERMEAS 0x37
#define CC_SETTEMPO     0x38

MidiListener::MidiListener(TimeBase* tb) :
    m_timebase(tb),
#ifdef USE_JACK_MIDI
    m_client(0),
    m_input(0)
#else
    m_seq(0),
    m_input(0)
#endif
{
#ifdef USE_JACK_MIDI
    if((m_client = jack_client_new("InConcert MIDI In")) == 0) {
        assert(false);
    }
    jack_set_process_callback(m_client, MidiListener::ProcessCallback, (void*)this);
    m_input = jack_port_register(
        m_client,
        "midi_in",
        JACK_DEFAULT_MIDI_TYPE,
        JackPortIsInput,
        0);
    jack_activate(m_client);
#else
    snd_seq_open(&m_seq, "default", SND_SEQ_OPEN_INPUT, 0);        
    m_input = snd_seq_create_simple_port(
        m_seq,
        "InConcert MIDI In",
        SND_SEQ_PORT_CAP_WRITE|SND_SEQ_PORT_CAP_SUBS_WRITE,
        SND_SEQ_PORT_TYPE_MIDI_GENERIC|SND_SEQ_PORT_TYPE_APPLICATION
        |SND_SEQ_PORT_TYPE_SOFTWARE);
#endif
}

MidiListener::~MidiListener()
{
#ifdef USE_JACK_MIDI
    jack_client_close(m_client);
#else
    snd_seq_delete_simple_port(m_seq, m_input);
#endif
}

#ifdef USE_JACK_MIDI
int MidiListener::ProcessCallback(jack_nframes_t nframes, void* arg)
{
    return static_cast<MidiListener*>(arg)->_ProcessCallback(nframes);
}
#endif

#ifdef USE_JACK_MIDI
int MidiListener::_ProcessCallback(jack_nframes_t nframes)
{
    void* port_buf = jack_port_get_buffer(m_input, nframes);
    jack_midi_event_t event;
    jack_nframes_t j;
    jack_nframes_t N = jack_midi_get_event_count(port_buf, nframes);
    for(j=0; j<N; ++j) {
        jack_midi_event_get(&event, port_buf, j, nframes);
        ProcessMidiEvent(event.buffer, event.size);
    }
    return 0;
}
#endif

#ifndef USE_JACK_MIDI
void MidiListener::run(void)
{
    snd_seq_event_t* ev;
    unsigned char buf[16];
    size_t len;
    while(snd_seq_event_input(m_seq, &ev) >=0) {
        // Only process control events
        if(ev->type == SND_SEQ_EVENT_CONTROLLER) {
            buf[0] = 0xB0 | ev->data.control.channel;
            buf[1] = ev->data.control.param;
            buf[2] = ev->data.control.value;
            len = 3;
            ProcessMidiEvent(buf, len);
        } else {
            cerr << "Event filtered out" << endl;
        }
    }
}
#endif

int MidiListener::ProcessMidiEvent(const unsigned char* buf, const size_t len)
{
    if(len == 0) return 0;
#ifndef NDEBUG
    for(size_t k=0; k<len; ++k) {
        cerr << hex << ((unsigned int)buf[k]) << " ";
    }
    cerr << endl;
#endif
    if( (len > 2)   // CC's only
        && (0xB0 == (buf[0] & 0xF0))
        && (buf[1] >= MIDI_CC_MIN)
        && (buf[1] <= MIDI_CC_MAX) ) {

        // Filter events
        if(buf[1] == CC_SET_CHANNEL) {
            cerr << "CC_SET_CHANNEL" << endl;
            if(buf[2] < 0x0F) {
                m_chan = buf[2];
                cerr << "Listening on channel " << hex << ((unsigned)m_chan) << endl;
            }
            return 0;
        }
        assert(m_chan <= 0x0F);
        if((buf[0] & 0x0F) != m_chan) {
            cerr << "Ignoring." << endl;
            return 0;
        }

        // Process the events that get through.
        switch(buf[1]) {
        case CC_REZERO:
            cerr << "CC_REZERO" << endl;
            m_timebase->re_zero();
            break;
        case CC_TAP:
            cerr << "CC_TAP" << endl;
            if(buf[2] & 0x40) m_timebase->tap();
            break;
        case CC_STARTSTOP:
            cerr << "CC_STARTSTOP" << endl;
            if(buf[2] & 0x40) {
                m_timebase->play();
            } else {
                m_timebase->stop();
            }
            break;
        case CC_INSERTORSKIP:
            cerr << "CC_INSERTORSKIP" << endl;
            if(buf[2] & 0x40) {
                m_timebase->skip_beat();
            } else {
                m_timebase->insert_beat();
            }
            break;
        case CC_TAPSREQUIRED:
            cerr << "CC_TAPSREQUIRED" << endl;
            if((buf[2] > 2)
               && (buf[2] <= TAPS_MAX)) {
                m_timebase->setTapsRequired((unsigned)buf[2]);
                cerr << "Now requiring " << m_timebase->getTapsRequired() << " taps" << endl;
            }
            break;
        case CC_TAPSCUECTL:
            cerr << "CC_TAPSCUECTL" << endl;
            if(buf[2] & 0x40) {
                m_timebase->setAutoStart(true);
            } else {
                m_timebase->setAutoStart(false);
            }
            break;
        case CC_BEATSPERMEAS:
            cerr << "CC_BEATSPERMEAS" << endl;
            if(buf[2]>0) m_timebase->setBeatsPerMeasure((unsigned)buf[2]);
            break;
        case CC_SETTEMPO:
            cerr << "CC_SETTEMPO" << endl;
            if(unsigned(buf[2])<= 127) m_timebase->setTempo(25.5 + double(buf[2])*1.50);
            break;
        default:
            assert(false);  // How did we get here... this event should have been filtered
        }
    }
    return 0;
}




