#ifndef __TAPTEMPO_HPP__
#define __TAPTEMPO_HPP__
/**********************************************-*- indent-tabs-mode:nil; -*-
 *                                                                         *
 *   InConcert - Tap Tempo Synchronization for Live Perforances            *
 *                                                                         *
 *   Copyright (C) 2008 by Gabriel M. Beddingfield                         *
 *   gabriel@teuton.org                                                    *
 *                                                                         *
 *   "For of him [God], and through him, and unto him, are all things.     *
 *   To him be the glory for ever. Amen." (Romans 11:36)                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; version 2 of the License, or any later  *
 *   version.                                                              *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/* TapTempo.hpp
   Gabriel M. Beddingfield
   2008-01-29

 */

#include <jack/types.h>
#include "Mutex.hpp"

using namespace std;

namespace InConcert
{

    // class TapTempo
    //
    // This class provides calculation of tempo and downbeat,
    // given a series of 2 or more taps.  See the tap() method.
    //
    class TapTempo
    {
    public:
        TapTempo();

        // Properties
        unsigned getTapsRequired(void) const { return m_taps_required; }
        void setTapsRequired(unsigned t);
        double getFrameRate(void) const { return m_frame_rate; }
        void setFrameRate(double rate);
        double getTimeout(void) const { return m_timeout; }
        void setTimeout(double t);
        void invalidate(void);
    
        // TapTempo::tap()
        //
        // Register a tap to define a new tempo and downbeat sync.  Supply
        // a frame number (tap_) and a reference to a double where the new
        // tempo may be written.
        //
        // Returns 0 and leaves beats_per_minute unchanged if we have not
        // reached the desired number of taps.  If this tap is the last
        // needed tap, it returns the frame number of the next beat, and
        // changes beats_per_minute to the new tempo.
        //
        // Future: Using the frame number is worthless unless the
        // transport is running.  Need to change that implementation so
        // that it can record valid taps even while the transport is stopped.
        //
        jack_nframes_t tap(jack_nframes_t tap_, double& beats_per_minute);
    
    private:
        // Mutex must already be locked before calling
        void invalidate_(void) {
            m_sum_taps = m_sum_i_taps = 0.0;
            m_taps_received = 0;
        }
    
        // Mutex must already be locked before calling
        jack_nframes_t lastTap(double& beats_per_minute);   
    
    private:
        double m_timeout; // seconds
        unsigned m_taps_required;
        unsigned m_taps_received;
        double m_last_tap;  // In seconds (frame/m_frame_rate)
        double m_sum_taps;
        double m_sum_i_taps;
        double m_frame_rate; // frames per second
        Mutex m_mutex;
    };
    
} // namespace InConcert
    
#endif // __TAPTEMPO_HPP__
