/**********************************************-*- indent-tabs-mode:nil; -*-
 *                                                                         *
 *   InConcert - Tap Tempo Synchronization for Live Perforances            *
 *                                                                         *
 *   Copyright (C) 2008 by Gabriel M. Beddingfield                         *
 *   gabriel@teuton.org                                                    *
 *                                                                         *
 *   "For of him [God], and through him, and unto him, are all things.     *
 *   To him be the glory for ever. Amen." (Romans 11:36)                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; version 2 of the License, or any later  *
 *   version.                                                              *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/* TimeBase.cpp
   Gabriel M. Beddingfield
   2008-01-29

   Implementation of TimeBase class.  See header for documentation.
 */

// Standard C++ libraries
#include <iostream>
#include <iomanip>
#include <stdexcept>
#include <cmath>
#include <cassert>

// POSIX libraries
#include <unistd.h>  // for usleep()

// Jack libraries
#include <jack/jack.h>
#include <jack/transport.h>
#include <jack/types.h>
// #include <alsa/asoundlib.h> // For midi with jack < 103

#include "TimeBase.hpp"
#include "MutexLocker.hpp"

//Darn!  Requires libjack >= 0.103.0, but etch uses 0.101
//#define JACK_DEFAULT_MIDI_TYPE "8 bit raw midi"

using namespace std;
using namespace InConcert;

void Epoch::init_frame_rates(jack_nframes_t rate) {
    m_anchor_beat_abs = (double)(m_bar-1.0)*m_beats_per_bar
        + (m_beat-1.0)
        + (double)m_tick/m_ticks_per_beat;
    m_frames_per_beat = (double)rate * 60.0 / m_beats_per_minute;
    m_frames_per_tick = m_frames_per_beat / m_ticks_per_beat;
    m_frames_per_measure = m_frames_per_beat * m_beats_per_bar;
}
    
void Epoch::set_jack_position_meta(jack_position_t* pos) {
    pos->valid = JackPositionBBT;
    pos->beats_per_bar = (int32_t)m_beats_per_bar;
    pos->beat_type = 4;
    pos->ticks_per_beat = (int32_t)m_ticks_per_beat;
    pos->beats_per_minute = m_beats_per_minute;
}
    
ostream& operator<<(ostream& os, Epoch x)
{
    os << "[" << x.m_bar
       << ":" << x.m_beat
       << ":" << x.m_tick
       << "] " << x.m_frame
       << " (" << x.m_anchor_beat_abs
       << ") @ " << x.m_frames_per_beat
       << " f/beat (" << x.m_beats_per_minute
       << " bpm) and |" << x.m_beats_per_bar
       << ":" << x.m_ticks_per_beat << "|";
    return os;
}
    
TimeBase::TimeBase(void) :
    m_state(Steady),
    m_state_mutex(0),
    m_time_reset(false),
    m_auto_start(false),
    m_client(0)
{
    m_state_mutex = new Mutex;

    // Initialize m_normal
    m_normal.m_frame = 0;
    m_normal.m_bar = 1;
    m_normal.m_beat = 1;
    m_normal.m_tick = 0;
    m_normal.m_beats_per_bar = 4;
    m_normal.m_ticks_per_beat = 360;
    m_normal.m_beats_per_minute = 120.0;
    m_ramp = m_normal;

    m_client = jack_client_open("TimeBase", JackNullOption, 0);
//     if(snd_rawmidi_open(&m_midi_in, 0, "virtual", 0)) {
//      cerr << "Couldn't open MIDI port" << endl;
//     }
    jack_activate(m_client);
    jack_set_timebase_callback(
        m_client,
        0,
        this->timebase,
        this);

    m_taps.setFrameRate((double)jack_get_sample_rate(m_client));
}

TimeBase::~TimeBase()
{
    jack_client_close(m_client);
}

// This should ONLY be called from TimeBase::timebase()
// otherwise, it would need to lock m_state_mutex.
void TimeBase::timebase_ramping(
        jack_transport_state_t state,
        jack_nframes_t nframes, 
        jack_position_t *pos,
        int new_pos)
{
    assert(pos->frame >= m_ramp.m_frame);
    assert(pos->frame < m_normal_queued.m_frame);
    assert(m_state == Ramping);
    //assert(m_state_mutex->status() == Mutex::Locked);

    if(new_pos || m_time_reset) {
        m_ramp.init_frame_rates(pos->frame_rate);
        m_ramp.set_jack_position_meta(pos);
        m_taps.setFrameRate((double)pos->frame_rate);
        m_time_reset = false;
    }
    m_ramp.getBBT(pos->frame, pos->bar, pos->beat, pos->tick);
}

// This should ONLY be called from TimeBase::timebase()
// otherwise, it would need to lock m_state_mutex.
void TimeBase::timebase_normal(
        jack_transport_state_t state,
        jack_nframes_t nframes, 
        jack_position_t *pos,
        int new_pos)
{
    // assert((m_state != Ramping) && (pos->frame > m_anchor_frame));
    //assert(m_state_mutex->status() == Mutex::Locked);

    if(new_pos || m_time_reset) {
        m_normal.init_frame_rates(pos->frame_rate);
        m_normal.set_jack_position_meta(pos);
        m_taps.setFrameRate((double)pos->frame_rate);
        m_time_reset = false;
    }
    m_normal.getBBT(pos->frame, pos->bar, pos->beat, pos->tick);
}

// Jack Timebase callback.
//
// This determines if we're in normal mode (Steady), ramping mode
// (Ramping), or if a ramp has been queued to start (RampQueued).  It
// calls the correct method depending on the state.
//
void TimeBase::timebase(
        jack_transport_state_t state,
        jack_nframes_t nframes, 
        jack_position_t *pos,
        int new_pos,
        void *arg)
{
    TimeBase* o = static_cast<TimeBase*>(arg);
    MutexLocker mutex(o->m_state_mutex);

    jack_nframes_t time = jack_frame_time(o->m_client);
    jack_nframes_t ctime = jack_frames_since_cycle_start(o->m_client);
#if 0
    cerr << "xport position: "
         << "usecs = " << pos->usecs
         << " frame = " << pos->frame
         << " jack_frame_time = " << time
         << " frame_time = " << pos->frame_time
         << " cyc_start_frames = " << ctime
         << endl;
#endif
    if(((o->m_state == Ramping)||(o->m_state == RampQueued))
       && (pos->frame >= o->m_ramp.m_frame)) {

        if(pos->frame < o->m_normal_queued.m_frame) {
            if(o->m_state == RampQueued) {
                o->m_state = Ramping;
                o->m_time_reset = true;
            }
            o->timebase_ramping(state, nframes, pos, new_pos);
        } else {
            cerr << "m_normal: " << o->m_normal << endl;
            cerr << "queued:   " << o->m_normal_queued << endl;
            o->m_normal = o->m_normal_queued;
            cerr << "Going steady: (nframes, rs, af) = ("
                 << pos->frame << ", "
                 << o->m_ramp.m_frame << ", "
                 << o->m_normal.m_frame
                 << ")" << endl;
            o->m_state = Steady;
            o->m_time_reset = true;
            o->timebase_normal(state, nframes, pos, new_pos);
        }
    } else {
        o->timebase_normal(state, nframes, pos, new_pos);
    }
}

// TapTempo has given us a frame to snap our beat to.  The trick
// to making this function work is to find the nearest beat to
// snap to, and to make sure that we don't skip any frames by
// accidentally setting m_ramp in the past.
void TimeBase::rampAndSnapTo(jack_nframes_t target, double new_tempo)
{
    Epoch ramp;
    Epoch normal = m_normal;
    double dummy;
    assert(JackTransportRolling == jack_transport_query(m_client, 0));
    jack_nframes_t frame_rate = jack_get_sample_rate(m_client);
    
    // Find the nearest BBT for our new normal anchor
    normal.getBBT(target, normal.m_bar, normal.m_beat, normal.m_tick);
    normal.m_frame = target;
    if((normal.m_ticks_per_beat-normal.m_tick) <= normal.m_tick) {
        ++normal.m_beat;
    }
    normal.m_tick = 0;
    normal.m_beats_per_minute = round(10.0*new_tempo)/10.0;
    normal.init_frame_rates(frame_rate);

    // Set up the ramp.  Don't start the ramp until the current
    // frame + max_buffer_size
    ramp = m_normal;
    ramp.m_frame = jack_get_buffer_size(m_client) + jack_get_current_transport_frame(m_client);
    assert(normal.m_frame > ramp.m_frame);
    m_normal.getBBT(ramp.m_frame, ramp.m_bar, ramp.m_beat, ramp.m_tick);
    ramp.init_frame_rates(frame_rate);
    assert(normal.m_anchor_beat_abs > ramp.m_anchor_beat_abs);
    ramp.m_beats_per_minute = 60.0*frame_rate*(normal.m_anchor_beat_abs - ramp.m_anchor_beat_abs)
        / (normal.m_frame - ramp.m_frame);
    ramp.init_frame_rates(frame_rate);

    cerr << "Old normal: " << m_normal << endl;
    cerr << "New normal: " << m_normal_queued << endl;
    cerr << "Ramp:       " << ramp << endl;
    m_normal_queued = normal;
    m_ramp = ramp;
    m_state = RampQueued;
}

// Register a tap for redefining the beat and tempo.
void TimeBase::tap(void) {
    double new_tempo;
    jack_nframes_t target;
    jack_nframes_t now;
    MutexLocker mutex(m_state_mutex);
    now = jack_frame_time(m_client);
    target = m_taps.tap(now, new_tempo);
    cerr << "Now = " << now << " Target = " << target << endl;
    if(target) {
        if(JackTransportRolling == jack_transport_query(m_client, 0)) {
            // Convert the target to transport frames (estimate).
            target = target - (jack_frame_time(m_client)
                               - jack_get_current_transport_frame(m_client));
            rampAndSnapTo(target, new_tempo);
        } else {
            m_normal.m_beats_per_minute = new_tempo;
            m_time_reset = true;
            m_auto_start = 1;
            if(m_auto_start) {
#               if(_XOPEN_SOURCE == 500)
                useconds_t usecs;
#               else
                unsigned long usecs;
#               endif
                jack_nframes_t jft = jack_frame_time(m_client);
                cerr << "target = " << dec << target << " frame_time = " << dec << jft << endl;
                #warning "Hardcoded 10*buffer_size... need to get this from the output port."
                // Estimate the number of microseconds from now that jack_frame_time()
                // will hit frame #target.
                // secs ~= (target-jft-latency)/samplerate
                // usecs = secs * 1000000
                usecs = floor(1000000.0*
                              double(target-jft-10*jack_get_buffer_size(m_client))
                              /double(jack_get_sample_rate(m_client)));
                cout << "usecs = " << dec << usecs;
                double totsleep = usecs;
                // usleep() over 1000000 usecs is not portable.
                if (usecs >= 1000000) {
                    cerr << "macro usecs" << endl;
                    unsigned secs = usecs / 1000000;
                    usecs = usecs % 1000000;
                    sleep(secs);
                    if(usecs) usleep(usecs);
                } else {
                    usleep(usecs);
                }
                jack_transport_start(m_client);
                cerr << "Slept for " << dec << totsleep << " for tempo " << new_tempo 
                     << " bpm (" << (60.0/new_tempo) << " sec/beat)" << endl;
                cerr << "I estimated latency at " << jack_get_buffer_size(m_client)*10 <<endl;
                cerr << "Jack reports port latencies as:" << endl;
                const char** ports = jack_get_ports(m_client, 0, 0, 0);
                jack_port_t* port;
                while (*ports) {
                    port = jack_port_by_name(m_client, *ports);
                    cerr << "    " << (*ports) << ": "
                         << "jack_port_get_latency()="
                         << jack_port_get_latency(port)
                         << "  jack_port_get_total_latency() = "
                         << jack_port_get_total_latency(m_client, port)
                         << endl;
                    ++ports;
                }
            }
        }
    }
}

// skip a beat by double-timing for one beat.
// Snaps to the downbeats.
//
// Create a new m_normal and m_ramp and queues the ramp (RampQueued).
//
void TimeBase::skip_beat(void)
{
    if(m_state == Ramping) {
        cout << "Already ramping... not skipping" << endl;
        return;
    }
    cout << "skipping beat..." << endl;
    jack_position_t pos;
    jack_transport_query(m_client, &pos);
    assert(pos.valid == JackPositionBBT);

    Epoch ramp, normal;
    ramp.m_beats_per_bar = (int32_t)pos.beats_per_bar;
    ramp.m_ticks_per_beat = (int32_t)pos.ticks_per_beat;
    if(pos.beat == ramp.m_beats_per_bar) {
        ramp.m_bar = pos.bar + 1;
        ramp.m_beat = 1;
    } else {
        ramp.m_bar = pos.bar;
        ramp.m_beat = pos.beat + 1;
    }
    ramp.m_tick = 0;

    normal = ramp;
    if(normal.m_beat == normal.m_beats_per_bar) {
        ++normal.m_bar;
        normal.m_beat = 1;
    } else {
        ++normal.m_beat;
    }

    MutexLocker mutex(m_state_mutex);
    ramp.m_frame = m_normal.getFrame(ramp.m_bar, ramp.m_beat, ramp.m_tick);
    normal.m_frame = m_normal.getFrame(normal.m_bar, normal.m_beat, normal.m_tick);
    ramp.m_beats_per_minute = 2.0 * pos.beats_per_minute;
    normal.m_beats_per_minute = pos.beats_per_minute;
    if(normal.m_beat == normal.m_beats_per_bar) {
        ++normal.m_bar;
        normal.m_beat = 1;
    } else {
        ++normal.m_beat;
    }
    normal.init_frame_rates(pos.frame_rate);
    ramp.init_frame_rates(pos.frame_rate);
    m_normal_queued = normal;
    m_ramp = ramp;
    m_state = RampQueued;

    // Ok, we set the ramp... now it's up to to
    // timebase() to do the Right Thing.
}

// skip a beat by half-timing for one beat.
// Snaps to the downbeats.
//
// Create a new m_normal and m_ramp and queues the ramp (RampQueued).
//
void TimeBase::insert_beat(void)
{
    if(m_state == Ramping) {
        cout << "Already ramping... not skipping" << endl;
        return;
    }
    cout << "inserting beat..." << endl;
    jack_position_t pos;
    jack_transport_query(m_client, &pos);
    assert(pos.valid == JackPositionBBT);

    Epoch ramp, normal;

    ramp.m_beats_per_bar = (int32_t)pos.beats_per_bar;
    ramp.m_ticks_per_beat = (int32_t)pos.ticks_per_beat;
    if(pos.beat == ramp.m_beats_per_bar) {
        ramp.m_bar = pos.bar + 1;
        ramp.m_beat = 1;
    } else {
        ramp.m_bar = pos.bar;
        ramp.m_beat = pos.beat + 1;
    }
    ramp.m_tick = 0;

    normal = ramp;
    if(normal.m_beat >= (normal.m_beats_per_bar-1)) {
        ++normal.m_bar;
        normal.m_beat = normal.m_beat + 2 - normal.m_beats_per_bar;
    } else {
        normal.m_beat += 2;
    }

    MutexLocker mutex(m_state_mutex);
    ramp.m_frame = m_normal.getFrame(ramp.m_bar, ramp.m_beat, ramp.m_tick);
    normal.m_frame = m_normal.getFrame(normal.m_bar, normal.m_beat, normal.m_tick);
    ramp.m_beats_per_minute = 0.5 * pos.beats_per_minute;
    normal.m_beats_per_minute = pos.beats_per_minute;
    if(normal.m_beat == 1) {
        --normal.m_bar;
        normal.m_beat = normal.m_beats_per_bar;
    } else {
        --normal.m_beat;
    }
    normal.init_frame_rates(pos.frame_rate);
    ramp.init_frame_rates(pos.frame_rate);
    m_normal_queued = normal;
    m_ramp = ramp;
    m_state = RampQueued;

    // Ok, we set the ramp... now it's up to to
    // timebase() to do the Right Thing.
}

void TimeBase::re_zero(void)
{
    MutexLocker mutex(m_state_mutex);
    if(m_state == Ramping || m_state == RampQueued) {
        m_state = Steady;
        m_normal = m_normal_queued;
    }
    MutexLocker e_mut(m_normal.m_mutex);
    m_state = Steady;
    m_normal.m_bar = 1;
    m_normal.m_beat = 1;
    m_normal.m_tick = 0;
    m_normal.m_frame = 0;
    m_time_reset = true;
    e_mut.unlock();
    mutex.unlock();
    jack_transport_locate(m_client, 0);    
}

void TimeBase::play(void)
{
    jack_transport_start(m_client);
}

void TimeBase::stop(void)
{
    jack_transport_stop(m_client);
}

unsigned TimeBase::getBeatsPerMeasure(void)
{
    MutexLocker mutex(m_state_mutex);
    unsigned rv = 0;
    if(m_state == Ramping) {
        rv = m_ramp.m_beats_per_bar;
    } else {
        assert((m_state == RampQueued) || (m_state == Steady));
        rv = m_normal.m_beats_per_bar;
    }
    return rv;
}

// When rolling.  This is basically a ramp-and-snap-to,
// taking effect on the next available beat (possibly beat
// 1 of a new measure if we reduce the beats_per_bar.
// In fact, I copied the rampAndSnapTo() code as a starting point.
// When stopped, it does the same, except there's no ramping,
// and we don't increment the beat by one.
void TimeBase::setBeatsPerMeasure(unsigned ct)
{
    Epoch normal = m_normal;
    MutexLocker mutex(m_state_mutex);
    jack_position_t pos;
    jack_transport_state_t trans_state = jack_transport_query(m_client, &pos);
    jack_nframes_t curframe = pos.frame;

    // Note:  Looping is deprecated in the API.
    if ((trans_state == JackTransportStarting) || (trans_state == JackTransportLooping)) {
        return;
    }

//     if (trans_state == JackTransportStopped) {
//         if(normal.m_beat >= ct) {
//             normal.m_beat = 1;
//         }
//         normal.m_frames_per_measure *= ((double)ct)/((double)normal.m_beats_per_bar));
//         normal.m_beats_per_bar = ct;
//         return;
//     }

    assert((trans_state == JackTransportRolling) || (trans_state == JackTransportStopped));
    bool rolling = (trans_state == JackTransportRolling);

    // Find the nearest BBT for our new normal anchor
    normal.getBBT(curframe, normal.m_bar, normal.m_beat, normal.m_tick);
    if (rolling) ++normal.m_beat; // next beat
    if(normal.m_beat > normal.m_beats_per_bar) {
        ++normal.m_bar;
        normal.m_beat = 1;
    }
    if (rolling) normal.m_tick = 0;
    if (rolling) {
        normal.m_frame = m_normal.getFrame(normal.m_bar, normal.m_beat, normal.m_tick);
        if(normal.m_beat > ct) {
            ++normal.m_bar;
            normal.m_beat = 1;
        }
    } else {
        normal.m_frame = curframe;
    }
    normal.m_beats_per_bar = ct;
    normal.init_frame_rates(jack_get_sample_rate(m_client));

    cerr << "Old normal: " << m_normal << endl;
    cerr << "New normal: " << normal << endl;
    if (rolling) {
        cerr << "Ramp:       " << normal << endl;
        m_normal_queued = normal;
        m_ramp = m_normal;
        m_state = RampQueued;
    } else {
        m_normal = normal;
    }
}

double TimeBase::getTempo(void)
{
    MutexLocker mutex(m_state_mutex);
    double rv = 0;
    if(m_state == Ramping) {
        rv = m_ramp.m_beats_per_minute;
    } else {
        assert((m_state == RampQueued) || (m_state == Steady));
        rv = m_normal.m_beats_per_minute;
    }
    return rv;
}

void TimeBase::setTempo(double bpm)
{
    assert(false); // not implemented yet.
}

Epoch::Epoch(void)
{
    m_mutex = new Mutex;
}

// Copy constructor... DO NOT COPY THE MUTEX!
// The copy constructor and assignment operator are needed
// because I wish to declare getBBT() and getFrame() as `const',
// but if I do, then I can't lock the mutex.  A workaround is
// to make the mutex a pointer.  But when the mutex becomes
// a pointer, we run the risk of double deletes and deadlocks
// when an Epoch instance is copied.
//
Epoch::Epoch(const Epoch& o) :
    m_frame(o.m_frame),
    m_bar(o.m_bar),
    m_beat(o.m_beat),
    m_tick(o.m_tick),
    m_beats_per_bar(o.m_beats_per_bar),
    m_ticks_per_beat(o.m_ticks_per_beat),
    m_beats_per_minute(o.m_beats_per_minute),
    m_frames_per_beat(o.m_frames_per_beat),
    m_frames_per_tick(o.m_frames_per_tick),
    m_frames_per_measure(o.m_frames_per_measure),
    m_anchor_beat_abs(o.m_anchor_beat_abs)
{
    m_mutex = new Mutex;
}

Epoch& Epoch::operator=(const Epoch& o)
{
    MutexLocker t_mx(m_mutex);
    MutexLocker o_mx(o.m_mutex);

    m_frame = o.m_frame;
    m_bar = o.m_bar;
    m_beat = o.m_beat;
    m_tick = o.m_tick;
    m_beats_per_bar = o.m_beats_per_bar;
    m_ticks_per_beat = o.m_ticks_per_beat;
    m_beats_per_minute = o.m_beats_per_minute;
    m_frames_per_beat = o.m_frames_per_beat;
    m_frames_per_tick = o.m_frames_per_tick;
    m_frames_per_measure = o.m_frames_per_measure;
    m_anchor_beat_abs = o.m_anchor_beat_abs;

    return *this;
}

Epoch::~Epoch()
{
    delete m_mutex;
}

// Epoch::getBBT()
//
// Calculates the bar:beat:tick of the given frame, based on the
// properties of the Epoch struct.
//
// Make sure that this can be called on itself.  i.e, that this code:
//
// Epoch t;
// t.getBBT(frame, t.bar, t.beat, t.tick);
//
// Does what you expect.
//
void Epoch::getBBT(jack_nframes_t frame,
                      int32_t& _bar,
                      int32_t& _beat,
                      int32_t& _tick) const
{
    double dbeat, bar, beat, tick;
    MutexLocker mutex(m_mutex);
    // Find the differenc ein beats from the anchor
    dbeat = ((double)frame - (double)m_frame)/m_frames_per_beat;
    // Now add in the anchor
    dbeat += m_anchor_beat_abs;
    assert(dbeat >= 0.0);
    // Now convert to bbt.
    tick = modf(dbeat, &beat) * m_ticks_per_beat;
    mutex.unlock();  // We're done with the data
    bar = beat/m_beats_per_bar; // floor later
    beat = fmod(beat, m_beats_per_bar);

    assert(bar >= 0.0);
    assert(beat >= 0.0);
    assert(tick >= 0.0);
    
    _bar = (int32_t)floor(bar+1.0);
    _beat = (int32_t)round(beat+1.0);  // beat should be nearly an integer
    _tick = (int32_t)floor(tick);
}

// Epoch::getFrame()
//
// Calculates the frame number of the given bar:beat:tick, based on
// the properties of the Epoch struct.
//
jack_nframes_t Epoch::getFrame(double bar, double beat, double tick) const
{
    MutexLocker mutex(m_mutex);
    // Calculate absolute beats (decimal w/ticks)
    double f = (bar-1.0)*m_beats_per_bar
        + (beat-1.0)
        + tick/m_ticks_per_beat;
    // offset by the anchor
    f -= m_anchor_beat_abs;
    // Convert to frames
    f *= m_frames_per_beat;
    // and offset by anchor
    f += m_frame;
    mutex.unlock();

    if( f < 0.0 ) {
        return (jack_nframes_t) 0;
    }
    return (jack_nframes_t) round(f);    
}
