#ifndef __TIMEBASE_HPP__
#define __TIMEBASE_HPP__
/**********************************************-*- indent-tabs-mode:nil; -*-
 *                                                                         *
 *   InConcert - Tap Tempo Synchronization for Live Performances           *
 *                                                                         *
 *   Copyright (C) 2008 by Gabriel M. Beddingfield                         *
 *   gabriel@teuton.org                                                    *
 *                                                                         *
 *   "For of him [God], and through him, and unto him, are all things.     *
 *   To him be the glory for ever. Amen." (Romans 11:36)                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; version 2 of the License, or any later  *
 *   version.                                                              *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/* TimeBase.cpp
   Gabriel M. Beddingfield
   2008-01-29

   Jack Transport master with tap-tempo sync.  For use with the Jack
   Audio Connection Toolkit.  Tested with jackd 0.101.1 on Debian 4.0
   (etch).

   The purpose of this program is to provide a Jack Transport
   synchronization for the purpose of live performance.  Using MIDI on
   stage often forces the band to follow the tempo of the sequencer.
   If a musician makes an error (e.g. a singer comes in 1 beat early),
   there is usually no means to recover.  This program is intended to
   fill that need for live performance.

   This program registers itself as a Jack Timebase Master.  Once the
   transport is started, Giving tempo taps will change the tempo, and
   ramp to synchronize the beat with the taps that were given.  It
   also has the ability to skip a beat (by changing the tempo to 2x
   for one beat), and to insert a beat (by changing the tempo to .5x
   for one beat).

   See below for build instructions.

   To understand the code, the main objects are these:

   main()   - a scaffolding for testing purposes.

   TimeBase - This is the Jack Timebase master.  It keeps track of
              where you are and makes the adjustments that you
              request.

   Epoch - This struct is only for use with the class TimeBase.
              It provides a reference point from which time is
              measured.  For example, it will know that frame
              562891 was bar:beat:tick = 10:2:0134, and the tempo
              there is 97.0 beats per minute.  It can calculate
              the frame of 11:1:0000 (or any bar:beat:tick) and
              and also the bar:beat:tick for any frame.

   TapTempo - This class manages the tap tempo feature.  It will
              accept 2 or more taps (configurable).  If more than two
              taps are given, the next beat and tempo are
              statistically determined (i.e. least squares) as opposed
              to averaging the taps (i.e. disregarding everything
              between the first and last tap).

   Known limitations:

   o If you change tempo, insert a beat, skip a beat, etc., then time
     0.00 will _not_ be bar 1, beat 1, tick 0.  It takes a special
     command to allow this to happen... and that command is not yet
     implemented.  :-P

   o It is assumed that you are making minor tempo corrections.  If
     you use the tap tempo to make drastic tempo corrections, it does
     not ramp smoothly.  For example, I've seen it dwell at 80 bpm
     when changing from 120 bpm to 180 bpm.

   o The taps don't work if the transport isn't running.  In fact, if
     you try it, you will probably crash the program (division by
     zero).

   o Yeah, tapping with the keyboard by typing T, <Enter> is
     wierd.  I intend to replace that with MIDI.
   
   Known Bugs:

   o When changing the tempo with TapTempo, there's a time between NOW
     and TimeBase::m_ramp.m_frame where we're using the NEW
     TimeBase::m_normal before we're supposed to.  This makes the tap
     tempo even jumpier.

   To-do:

   o Divide up the files like a normal project.  :-)

   o Implement the stuff mentioned above as "I intend to..."

   o Change the Mutex to a ReadWriteMutex (many readers, one writer).
     The Mutex code is currently the bottleneck.

   o Some private methods require that the mutex already be locked.
     There's no good way to test that... and it just doesn't seem
     right.  There's no protection against some rogue programmer (me)
     from using these methods without checking the header to see
     "mutex must be locked."

   o Implement the 'const' keyword on class methods where appropriate.
     Note that this typically requires that the Mutex in the class be
     a pointer rather than a heavy object -- which makes copy
     constructors more complicated.

   o Better debugging message scheme.

   BUILD INSTRUCTIONS
   ==================

   Library requirements:
     o C Standard Libraries
     o C++ Standard Libraries
     o POSIX libraries (especially pthread)
     o Jack Audio Connection Kit library

   Compiling:

   g++ -o inconcert inconcert.cpp -ljack -lpthread

 */

// Standard C++ libraries
#include <iostream>
#include <iomanip>
#include <cstdio>
#include <memory>
#include <stdexcept>
#include <cmath>
#include <cassert>

// POSIX libraries
#include <pthread.h>
#include <unistd.h> // sleep()

// Jack libraries
#include <jack/jack.h>
#include <jack/transport.h>
#include <jack/types.h>
// #include <alsa/asoundlib.h> // For midi with jack < 103

#include "TapTempo.hpp"

//Darn!  Requires libjack >= 0.103.0, but etch uses 0.101
//#define JACK_DEFAULT_MIDI_TYPE "8 bit raw midi"

namespace InConcert
{
    class Mutex;
    class TapTempo;

    // struct Epoch
    //
    // This is a convenience class for the TimeBase class.
    // It provides an anchor point and BBT calculation.
    // This allows the TimeBase class to prepare new anchor
    // points and quickly switch between them.
    //
    struct Epoch
    {
        jack_nframes_t m_frame;
        int32_t m_bar,
            m_beat,
            m_tick,
            m_beats_per_bar,
            m_ticks_per_beat;
        double m_beats_per_minute;
        // Metadata to speed computation:
        double m_frames_per_beat,
            m_frames_per_tick,
            m_frames_per_measure,
            m_anchor_beat_abs;  // first beat is 0.0
        Mutex* m_mutex;
    
        Epoch(void);
        ~Epoch();
        // Make sure mutex is not copied:
        Epoch(const Epoch& o);
        Epoch& operator=(const Epoch& o);

        void init_frame_rates(jack_nframes_t rate);    
        void set_jack_position_meta(jack_position_t* pos);    

        void getBBT(jack_nframes_t fr,
                    int32_t& bar,
                    int32_t& beat,
                    int32_t& tick) const;    
        jack_nframes_t getFrame(double bar,
                                double beat,
                                double tick) const;
    };
    
    
    // class TimeBase
    //
    // This class acts as a Jack Transport Master for synchronizing
    // the computer audio with a live ensemble during the perfomance.
    // It features a Tap Tempo interface and an interface to insert
    // or skip a beat.
    //
    class TimeBase
    {
    public:
        // Steady = m_normal is the current time base.
        //
        // Ramping = m_ramp is the current time base, until we reach
        //           m_normal.m_frame.
        //
        // RampQueued = when m_ramp.m_frame is reached, switch over to
        //              Ramping.
        //
        typedef enum { Steady, Ramping, RampQueued } timebase_state;
    
    private:
        timebase_state m_state;
        Mutex* m_state_mutex;
        // m_normal_queued is to hold the new m_normal, to be
        // applied _after_ ramping.
        Epoch m_normal, m_ramp, m_normal_queued;
        TapTempo m_taps;
        bool m_time_reset;  // Reset the timebase now.
        bool m_auto_start;

        // Jack stuff
        jack_client_t* m_client;
        // snd_rawmidi_t* m_midi_in;
    
    public:
        TimeBase();
        ~TimeBase();
    
        timebase_state getState(void) const { return m_state; }
    
        // Tap settings
        unsigned getTapsRequired(void) const { return m_taps.getTapsRequired(); }
        void setTapsRequired(unsigned ct) { m_taps.setTapsRequired(ct); }
        double getTapsTimeout(void) const { return m_taps.getTimeout(); }
        void setTapsTimeout(double t) { m_taps.setTimeout(t); }
    
        // methods
    
        // Methods needed by user
        void tap(void);
        void skip_beat(void);
        void insert_beat(void);
        void re_zero(void);
        void play(void);
        void stop(void);
        void setAutoStart(bool state) {
            m_auto_start = state;
        }
        bool getAutoStart(void) { return m_auto_start; }
        unsigned getBeatsPerMeasure(void);
        void setBeatsPerMeasure(unsigned ct);
        double getTempo(void);
        void setTempo(double bpm);

        // Jack timebase callback, routes to the private timebase_normal()
        // and timebase_ramping methods.
        static void timebase(
            jack_transport_state_t state,
            jack_nframes_t nframes, 
            jack_position_t *pos,
            int new_pos,
            void *arg);
    
    private:
        // Jack timebase callback when ramping
        void timebase_ramping(
            jack_transport_state_t state,
            jack_nframes_t nframes, 
            jack_position_t *pos,
            int new_pos);
    
        // Jack timebase callback when in normal (Steady) mode.
        void timebase_normal(
            jack_transport_state_t state,
            jack_nframes_t nframes, 
            jack_position_t *pos,
            int new_pos);
    
        // Called after all the taps have been given to m_taps.
        //
        // Mutex must already be locked before calling this
        void rampAndSnapTo(jack_nframes_t target, double new_tempo);
    };

} // namespace InConcert

ostream& operator<<(ostream& os, InConcert::Epoch x);
    
#endif // __TIMEBASE_HPP__
