/**********************************************-*- indent-tabs-mode:nil; -*-
 *                                                                         *
 *   InConcert - Tap Tempo Synchronization for Live Perforances            *
 *                                                                         *
 *   Copyright (C) 2008 by Gabriel M. Beddingfield                         *
 *   gabriel@teuton.org                                                    *
 *                                                                         *
 *   "For of him [God], and through him, and unto him, are all things.     *
 *   To him be the glory for ever. Amen." (Romans 11:36)                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; version 2 of the License, or any later  *
 *   version                                                               *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/* main.cpp
   Gabriel M. Beddingfield
   2008-01-29

   Jack Transport master with tap-tempo sync.  For use with the Jack
   Audio Connection Toolkit.  Tested with jackd 0.101.1 on Debian 4.0
   (etch).

   The purpose of this program is to provide a Jack Transport
   synchronization for the purpose of live performance.  Using MIDI on
   stage often forces the band to follow the tempo of the sequencer.
   If a musician makes an error (e.g. a singer comes in 1 beat early),
   there is usually no means to recover.  This program is intended to
   fill that need for live performance.

   This program registers itself as a Jack Timebase Master.  Once the
   transport is started, Giving tempo taps will change the tempo, and
   ramp to synchronize the beat with the taps that were given.  It
   also has the ability to skip a beat (by changing the tempo to 2x
   for one beat), and to insert a beat (by changing the tempo to .5x
   for one beat).

   See below for build instructions.

   To understand the code, the main objects are these:

   main()   - a scaffolding for testing purposes.

   TimeBase - This is the Jack Timebase master.  It keeps track of
              where you are and makes the adjustments that you
              request.

   Epoch - This struct is only for use with the class TimeBase.
              It provides a reference point from which time is
              measured.  For example, it will know that frame
              562891 was bar:beat:tick = 10:2:0134, and the tempo
              there is 97.0 beats per minute.  It can calculate
              the frame of 11:1:0000 (or any bar:beat:tick) and
              and also the bar:beat:tick for any frame.

   TapTempo - This class manages the tap tempo feature.  It will
              accept 2 or more taps (configurable).  If more than two
              taps are given, the next beat and tempo are
              statistically determined (i.e. least squares) as opposed
              to averaging the taps (i.e. disregarding everything
              between the first and last tap).

   Known limitations:

   o If you change tempo, insert a beat, skip a beat, etc., then time
     0.00 will _not_ be bar 1, beat 1, tick 0.  It takes a special
     command to allow this to happen... and that command is not yet
     implemented.  :-P

   o It is assumed that you are making minor tempo corrections.  If
     you use the tap tempo to make drastic tempo corrections, it does
     not ramp smoothly.  For example, I've seen it dwell at 80 bpm
     when changing from 120 bpm to 180 bpm.

   o The taps don't work if the transport isn't running.  In fact, if
     you try it, you will probably crash the program (division by
     zero).

   o Yeah, tapping with the keyboard by typing T, <Enter> is
     wierd.  I intend to replace that with MIDI.
   
   Known Bugs:

   o When changing the tempo with TapTempo, there's a time between NOW
     and TimeBase::m_ramp.m_frame where we're using the NEW
     TimeBase::m_normal before we're supposed to.  This makes the tap
     tempo even jumpier.

   To-do:

   o Divide up the files like a normal project.  :-)

   o Implement the stuff mentioned above as "I intend to..."

   o Change the Mutex to a ReadWriteMutex (many readers, one writer).
     The Mutex code is currently the bottleneck.

   o Some private methods require that the mutex already be locked.
     There's no good way to test that... and it just doesn't seem
     right.  There's no protection against some rogue programmer (me)
     from using these methods without checking the header to see
     "mutex must be locked."

   o Implement the 'const' keyword on class methods where appropriate.
     Note that this typically requires that the Mutex in the class be
     a pointer rather than a heavy object -- which makes copy
     constructors more complicated.

   o Better debugging message scheme.

   BUILD INSTRUCTIONS
   ==================

   Library requirements:
     o C Standard Libraries
     o C++ Standard Libraries
     o POSIX libraries (especially pthread)
     o Jack Audio Connection Kit library

   Compiling:

   g++ -o inconcert inconcert.cpp -ljack -lpthread

 */

// Standard C++ libraries
#include <iostream>
#include "TimeBase.hpp"
#include "MidiListener.hpp"
#include "InConcertDlg.hpp"
#include <QApplication>

using namespace std;
using namespace InConcert;

// main()
//
// A scaffolding to test the code.  Note that some debugging messages
// will be dumped to your screen... but you can still enter a command
// independant of this.
//
// Commands:
//
// f - skip forward by one beat.
// b - insert a beat ("back")
// t - enter a "tap" for the tap tempo feature.
// q - quit.
//
// You will have to follow each command by pressing <Enter>
//
int main(int argc, char* argv[])
{
    QApplication app(argc, argv);
    TimeBase tb;
    MidiListener ml(&tb);
    InConcertDlg c(0, &tb);

    ml.start();
    c.show();
    ml.terminate();

    return app.exec();
}

